import requests
import json
from parsons import Table


class BillCom(object):
    """
    `Args:`
        user_name: str
            The Bill.com username
        password: str
            The Bill.com password
        org_id: str
            The Bill.com organization id
        dev_key: str
            The Bill.com dev key
        api_url:
            The Bill.com end point url
    """

    def __init__(self, user_name, password, org_id, dev_key, api_url):
        self.headers = {
            "Content-Type": "application/x-www-form-urlencoded"
        }
        params = {
            "userName": user_name,
            "password": password,
            "orgId": org_id,
            "devKey": dev_key
        }
        response = requests.post(url="%sLogin.json" % api_url,
                                 data=params,
                                 headers=self.headers)
        self.dev_key = dev_key
        self.api_url = api_url
        self.session_id = response.json()['response_data']['sessionId']

    def _get_payload(self, data):
        """
        `Args:`
            data: dict
                A dictionary containing the payload to be sent in the request.
                The dev_key and sessionId should not be included as they are
                dealt with separately.

        `Returns:`
            A dictionary of the payload to be sent in the request with the
            dev_key and sessionId added.
        """
        return {"devKey": self.dev_key,
                "sessionId": self.session_id,
                "data": json.dumps(data)}

    def _post_request(self, data, action, object_name):
        """
        `Args:`
            data: dict
                A dictionary containing the payload to be sent in the request.
                The dev_key and sessionId should not be included as they are
                dealt with separately.
            action: str
                The action to be taken on the given object.
                Possible values are "List", "Read", "Create", and "Send".
            object_name: str
                The id of the object. Possible values are "User", "Customer", and "Invoice".

        `Returns:`
            A dictionary containing the JSON response from the post request.
        """

        if action == 'Read':
            url = "%sCrud/%s/%s.json" % (self.api_url, action, object_name)
        elif action == 'Create':
            data['obj']['entity'] = object_name
            url = "%sCrud/%s/%s.json" % (self.api_url, action, object_name)
        elif action == "Send":
            url = "%s%s%s.json" % (self.api_url, action, object_name)
        else:
            url = "%s%s/%s.json" % (self.api_url, action, object_name)
        payload = self._get_payload(data)
        response = requests.post(url=url, data=payload, headers=self.headers)
        return response.json()

    def _get_request_response(self, data, action, object_name, field="response_data"):
        """
        `Args:`
            data: dict
                A dictionary containing the payload to be sent in the request.
                The dev_key and sessionId should not be included as they are
                dealt with separately.
            action: str
                The action to be taken on the given object.
                Possible values are "List", "Read", "Create", and "Send".
            object_name: str
                The id of the object. Possible values are "User", "Customer", and "Invoice".
            field: str
                The JSON field where the response data is stored. Defaults to "response_data".

        `Returns:`
            A dictionary containing the choosen field from the JSON response from the post request.
        """
        r = self._post_request(data, action, object_name)[field]

        if action == "List":
            return self._paginate_list(r, data, object_name)

        return r

    def _paginate_list(self, response, data, object_name, field="response_data"):
        """
        Internal method to paginate through and concatenate results of lists larger than max
        `Args:`
            response: list of dicts
                Data from an initial list call
            data: dict
                Start, max, and kwargs from initial list call
            object_name: str
                Name of the object being listed
        """

        r_table = Table(response)
        max_ct = data['max']

        # Flag to simulate do-while
        first_run = True

        while first_run or len(response) == max_ct:
            first_run = False
            data['start'] += max_ct
            response = self._post_request(data, "List", object_name)[field]
            r_table.concat(Table(response))

        return r_table

    def get_user_list(self, start_user=0, max_user=999, **kwargs):
        """
        `Args:`
            start_user: int
                The index of first user to return. Starts from 0 (not 1).
            max_user: str
                The index of the max user to return
            **kwargs:
                Any other fields to pass

        `Returns:`
            A Parsons Table of user information for every user from start_user to max_user.
        """
        data = {
           "start": start_user,
           "max": max_user,
           **kwargs
        }

        return self._get_request_response(data, "List", "User")

    def get_customer_list(self, start_customer=0, max_customer=999, **kwargs):
        """
        `Args:`
            start_customer: int
                The index of first customer to return. Starts from 1 (not 0).
            max_customer: str
                The index of the max customer to return
            **kwargs:
                Any other fields to pass

        `Returns:`
            A Parsons Table of customer information for every user from start_customer
            to max_customer.
        """
        data = {
           "start": start_customer,
           "max": max_customer,
           **kwargs
        }

        return self._get_request_response(data, "List", "Customer")

    def get_invoice_list(self, start_invoice=0, max_invoice=999, **kwargs):
        """
        `Args:`
            start_invoice: int
                The index of first customer to return. Starts from 1 (not 0).
            max_invoice: str
                The index of the max customer to return
            **kwargs:
                Any other fields to pass

        `Returns:`
            A list of dictionaries of invoice information for every invoice from start_invoice
            to max_invoice.
        """
        data = {
           "start": start_invoice,
           "max": max_invoice,
           **kwargs
        }

        return self._get_request_response(data, "List", "Invoice")

    def read_customer(self, customer_id):
        """
        `Args:`
            customer_id: str
                The id of the customer to query

        `Returns:`
            A dictionary of the customer's information.
        """
        data = {
            'id': customer_id
        }
        return self._get_request_response(data, "Read", "Customer")

    def read_invoice(self, invoice_id):
        """
        `Args:`
            invoice_id: str
                The id of the invoice to query

        `Returns:`
            A dictionary of the invoice information.
        """
        data = {
           "id": invoice_id
        }
        return self._get_request_response(data, "Read", "Invoice")

    def check_customer(self, customer1, customer2):
        """
        `Args:`
            customer1: dict
                A dictionary of data on customer1
            customer2: dict
                A dictionary of data on customer2

        `Returns:`
            True if either
                1. customer1 and customer2 have the same id
                OR
                2. customer1 has no id and customer1 customer2 have the same email address
            False otherwise

        """
        if "id" in customer1.keys():
            if customer1["id"] == customer2["id"]:
                return True
        if "id" not in customer1.keys() and customer2['email']:
            if customer1['email'].lower() == customer2['email'].lower():
                return True
        return False

    def get_or_create_customer(self, customer_name, customer_email, **kwargs):
        """
        `Args:`
            customer_name: str
                The name of the customer
            customer_email: str
                The customer's email
        `Keyword Args:`
            **kwargs:
                Any other fields to store about the customer.

        `Returns:`
            A dictionary of the customer's information including an id.
            If the customer already exists, this function will not
            create a new id and instead use the existing id.
        """
        customer = {"name": customer_name,
                    "email": customer_email,
                    **kwargs}

        # check if customer already exists
        customer_list = self.get_customer_list()
        for existing_customer in customer_list:
            if self.check_customer(customer, existing_customer):
                return existing_customer
        # customer doesn't exist, create
        data = {
            "obj": customer
        }
        return self._get_request_response(data, "Create", "Customer")

    def create_invoice(self, customer_id, invoice_number, invoice_date,
                       due_date, invoice_line_items, **kwargs):
        """
        `Args:`
            customer_id: str
                The customer's id
            invoice_number: str
                The invoice number. Every invoice must have a distinct
                invoice number.
            invoice_date: str
                The invoice date. This can be the date the invoice was
                generated of any other relevant date.
            due_date: str
                The date on which the invoice is due.
            invoice_line_items: list
                A list of dicts, one for each line item in the invoice.
                The only required field is "quantity".
            **kwargs:
                Any other invoice details to pass.

        `Returns:`
            A dictionary of the invoice's information including an id.
        """
        for invoice_line_item in invoice_line_items:
            if "entity" not in invoice_line_item:
                invoice_line_item['entity'] = 'InvoiceLineItem'
        data = {
            "obj": {"customerId": customer_id,
                    "invoiceNumber": invoice_number,
                    "invoiceDate": invoice_date,
                    "dueDate": due_date,
                    "invoiceLineItems": invoice_line_items,
                    **kwargs
                    }
        }
        return self._get_request_response(data, "Create", "Invoice")

    def send_invoice(self, invoice_id, from_user_id, to_email_addresses,
                     message_subject, message_body, **kwargs):
        """
        `Args:`
            invoice_id: str
                The id of the invoice to send
            from_user_id: str
                The id of the Bill.com user from whom to send the email
            to_email_addresses:
                The customer's email address
            message_subject:
                The subject of the email to send to the customer
            message_body:
                The body of the email to send to the customer
            **kwargs:
                Any other details for sending the invoice

        `Returns:`
            A dictionary of the sent invoice.
        """
        data = {
          "invoiceId": invoice_id,
          "headers": {
            "fromUserId": from_user_id,
            "toEmailAddresses": to_email_addresses,
            "subject": message_subject,
            **kwargs
          },
          "content": {
            "body": message_body
          }
        }
        return self._get_request_response(data, "Send", "Invoice")



import datetime
import logging
import petl
import re
import requests
import time

from dateutil.parser import parse as parse_date

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S UTC'
"""Datetime format for sending date's to the API."""

REQUEST_HEADERS = {
    # For some reason, RTV's firewall REALLY doesn't like the
    # user-agent string that Python's request library gives by default,
    # though it seems fine with the curl user agent
    # For more info on user agents, see:
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/User-Agent
    'user-agent': 'curl/7.54.0'
}
"""Standard request header for sending requests to the API."""

STATUS_URL_PARSE_REGEX = re.compile(r'(\d+)$')
"""Regex for parsing the report ID from the status URL."""


class RTVFailure(Exception):
    """Exception raised when there is an error with the connector."""


class RockTheVote:
    """
    `Args:`
        partner_id: str
            The RockTheVote partner ID for the RTV account
        partner_api_key: str
            The API Key for the partner
    `Returns`:
        RockTheVote class
    """

    def __init__(self, partner_id=None, partner_api_key=None):
        self.partner_id = check_env.check('RTV_PARTNER_ID', partner_id)
        self.partner_api_key = check_env.check('RTV_PARTNER_API_KEY', partner_api_key)

        self.client = APIConnector('https://vr.rockthevote.com/api/v4', headers=REQUEST_HEADERS)

    def create_registration_report(self, before=None, since=None):
        """
        Create a new registration report.

        `Args:`
            before: str
                Date before which to return registrations for
            since: str
                Date for filtering registrations
        `Returns:`
            int
                The ID of the created report.
        """
        report_url = f'{self.client.uri}/registrant_reports.json'
        # Create the report for the new data
        report_parameters = {
            'partner_id': self.partner_id,
            'partner_API_key': self.partner_api_key,
        }

        if since:
            since_date = parse_date(since)
            report_parameters['since'] = since_date.strftime(DATETIME_FORMAT)
        if before:
            before_date = parse_date(before)
            report_parameters['before'] = before_date.strftime(DATETIME_FORMAT)

        # The report parameters get passed into the request as JSON in the body
        # of the request.
        response = self.client.request(report_url, 'post', json=report_parameters)
        if response.status_code != requests.codes.ok:
            print(response.text)
            raise RTVFailure("Couldn't create RTV registrations report")

        response_json = response.json()
        # The RTV API says the response should include the report_id, but I have not found
        # that to be the case
        report_id = response_json.get('report_id')
        if report_id:
            return report_id

        # If the response didn't include the report_id, then we will parse it out of the URL.
        status_url = response_json.get('status_url')
        url_match = STATUS_URL_PARSE_REGEX.search(status_url)
        if url_match:
            report_id = url_match.group(1)

        return report_id

    def get_registration_report(self, report_id, block=False, poll_interval_seconds=60,
                                report_timeout_seconds=3600):
        """
        Get data from an existing registration report.

        `Args:`
            report_id: int
                The ID of the report to get data from
            block: bool
                Whether or not to block execution until the report is complete
            poll_interval_seconds: int
                If blocking, how long to pause between attempts to check if the report is done
            report_timeout_seconds: int
                If blocking, how long to wait for the report before timing out
        `Returns:`
            Parsons Table
                Parsons table with the report data.
        """
        credentials = {
            'partner_id': self.partner_id,
            'partner_API_key': self.partner_api_key,
        }
        status_url = f'{self.client.uri}/registrant_reports/{report_id}'
        download_url = None

        # Let's figure out at what time should we just give up because we waited
        # too long
        end_time = datetime.datetime.now() + datetime.timedelta(seconds=report_timeout_seconds)

        # If we have a download URL, we can move on and just download the
        # report. Otherwise, as long as we haven't run out of time, we will
        # check the status.
        while not download_url and datetime.datetime.now() < end_time:
            logger.debug(f'Registrations report not ready yet, sleeping %s seconds',
                         poll_interval_seconds)

            # Check the status again via the status endpoint
            status_response = self.client.request(status_url, 'get', params=credentials)

            # Check to make sure the call got a valid response
            if status_response.status_code == requests.codes.ok:
                status_json = status_response.json()

                # Grab the download_url from the response.
                download_url = status_json.get('download_url')

                if not download_url and not block:
                    return None
            else:
                raise RTVFailure("Couldn't get report status")

            if not download_url:
                # We just got the status, so we should wait a minute before
                # we check it again.
                time.sleep(poll_interval_seconds)

        # If we never got a valid download_url, then we timed out waiting for
        # the report to generate. We will log an error and exit.
        if not download_url:
            raise RTVFailure('Timed out waiting for report')

        # Download the report data
        download_response = self.client.request(download_url, 'get', params=credentials)

        # Check to make sure the call got a valid response
        if download_response.status_code == requests.codes.ok:
            report_data = download_response.text

            # Load the report data into a Parsons Table
            table = Table.from_csv_string(report_data)

            # Transform the data from the report's CSV format to something more
            # Pythonic (snake case)
            normalized_column_names = [
                re.sub(r'\s', '_', name).lower()
                for name in table.columns
            ]
            normalized_column_names = [
                re.sub(r'[^A-Za-z\d_]', '', name)
                for name in normalized_column_names
            ]
            table.table = petl.setheader(table.table, normalized_column_names)
            return table
        else:
            raise RTVFailure('Unable to download report data')

    def run_registration_report(self, before=None, since=None, poll_interval_seconds=60,
                                report_timeout_seconds=3600):
        """
        Run a new registration report.

        This method will block until the report has finished generating, or until the specified
        timeout is reached.

        `Args:`
            before: str
                Date before which to return registrations for
            since: str
                Date for filtering registrations
            poll_interval_seconds: int
                If blocking, how long to pause between attempts to check if the report is done
            report_timeout_seconds: int
                If blocking, how long to wait for the report before timing out
        `Returns:`
            int
                The ID of the created report.
        """
        report_id = self.create_registration_report(before=before, since=since)
        return self.get_registration_report(report_id, block=True,
                                            poll_interval_seconds=poll_interval_seconds,
                                            report_timeout_seconds=report_timeout_seconds)



from requests import request
import time
from parsons.etl import Table
from parsons.utilities import check_env
import logging

logger = logging.getLogger(__name__)

CT_URI = "https://api.crowdtangle.com/"
PAGE_SIZE = 100
REQUEST_SLEEP = 10  # CT has a rather agressive 6 requests per minute rate limit.


class CrowdTangle(object):
    """
    Instantiate CrowdTangle Class

    `Args:`
        api_key: str
            A valid CrowdTangle API key. Not required if ``CROWDTANGLE_API_KEY`` env
            variable set.
    `Returns:`
        CrowdTangle Class
    """

    def __init__(self, api_key=None):

        self.api_key = check_env.check('CT_API_KEY', api_key)
        self.uri = CT_URI

    def base_request(self, endpoint, req_type='GET', args=None):
        # Internal Request Method

        url = f'{self.uri}/{endpoint}'
        base_args = {'token': self.api_key,
                     'count': PAGE_SIZE}

        # Add any args passed through to the base args
        if args is not None:
            base_args.update(args)

        r = request(req_type, url, params=base_args).json()
        json = r['result']
        keys = list(json.keys())
        data = json[keys[0]]

        while 'nextPage' in list(json['pagination'].keys()):
            logger.info(f"Retrieving {PAGE_SIZE} rows.")
            time.sleep(REQUEST_SLEEP)
            next_url = json['pagination']['nextPage']
            r = request(req_type, next_url).json()
            json = r['result']
            data.extend(json[keys[0]])

        logger.info(f"Retrieved {len(data)} rows.")

        return data

    def base_unpack(self, ParsonsTable):
        # Unpack base request

        logger.debug("Working to unpack the Parsons Table...")
        logger.debug(f'Starting with {len(ParsonsTable.columns)} columns...')
        sample = ParsonsTable[0]

        col_dict = {}
        for x in ParsonsTable.columns:
            col_dict[x] = str(type(sample[x]))

        for col in col_dict:
            if col_dict[col] == "<class 'dict'>":
                ParsonsTable.unpack_dict(col)
            elif col_dict[x] == "<class 'list'>":
                ParsonsTable.unpack_list(col)

        logger.info(f'There are now {len(ParsonsTable.columns)} columns...')
        return ParsonsTable

    def unpack(self, ParsonsTable):

        if ParsonsTable.num_rows == 0:
            return None

        first_count = len(ParsonsTable.columns)
        second_count = first_count + 1

        while second_count > first_count:
            first_count = len(ParsonsTable.columns)
            self.base_unpack(ParsonsTable)
            second_count = len(ParsonsTable.columns)

        return ParsonsTable

    def list_to_string(self, list_arg):

        if list_arg:
            return ','.join(list_arg)
        else:
            return None

    def get_posts(self, start_date=None, end_date=None, language=None, list_ids=None,
                  min_interations=None, search_term=None, types=None):
        """
        Return advocates (person records).

        `Args:`
            start_date: str
                Filter to the earliest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss``).
            end_date: str
                Filter to the latest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss``).
            language: str
                Filter to 2-character Locale code. Some languages require more
                than two characters: Chinese (Simplified) is zh-CN and
                Chinese (Traditional) is zh-TW.
            list_ids: list
                Filter to the ids of lists or saved searches to retrieve.
            min_interactions: int
                Filter to posts with total interactions above this threshold.
            search_team: str
                Returns only posts that match this search term. For multiple terms, separate
                with commas for OR, use quotes for phrases.
            types: list
                Filter to post types including:
                * ``episode``
                * ``extra_clip``
                * ``link``
                * ``live_video``
                * ``live_video_complete``
                * ``live_video_scheduled``
                * ``native_video``
                * ``photo``
                * ``status``
                * ``trailer``
                * ``tweet``
                * ``vimeo``
                * ``vine``
                * ``youtube``

                If you want all live videos (whether currently or formerly live),
                pass include both ``live_video`` and ``live_video_complete``
                parameters.

                The ``video`` type does not mean all videos, it refers to videos
                that are not ``native_video``, ``youtube`` or ``vine``.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {'startDate': start_date,
                'endDate': end_date,
                'language': language,
                'listIds': self.list_to_string(list_ids),
                'minInteractions': min_interations,
                'searchTerm': search_term,
                'types': types}

        logger.info("Retrieving posts.")
        pt = Table(self.base_request('posts', args=args))
        logger.info(f'Retrieved {pt.num_rows} posts.')
        self.unpack(pt)
        return pt

    def get_leaderboard(self, start_date=None, end_date=None, list_ids=None, account_ids=None):
        """
        Return advocates (person records).

        `Args:`
            start_date: str
                Filter to the earliest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss``).
            end_date: str
                Filter to the latest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss``).
            list_ids: list
                Filter to the ids of lists or saved searches to retrieve.
            account_ids: list
                A list of CrowdTangle accountIds to retrieve leaderboard data for.
                This and ``list_id`` are mutually exclusive; if both are sent, the
                ``account_ids`` value will be used.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {'startDate': start_date,
                'endDate': end_date,
                'listIds': self.list_to_string(list_ids),
                'accountIds': self.list_to_string(account_ids)}

        pt = Table(self.base_request('leaderboard', args=args))
        logger.info(f'Retrieved {pt.num_rows} records from the leaderbooard.')
        self.unpack(pt)
        return pt

    def get_links(self, link, start_date=None, end_date=None, include_summary=None,
                  platforms=None):
        """
        Return posts based on a specific link.

        `Args:`
            link: str
                The link to filter posts to.
            start_date: str
                Filter to the earliest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss``).
            end_date: str
                Filter to the latest date at which a post could be posted.
                The time is formatted as UTC (e.g. ``yyyy-mm-ddThh:mm:ss``).
            include_summary: boolean
                Adds a ``summary`` column with account statistics for each platform
                that has posted this link. It will look beyond the count
                requested to summarize across the time searched.
                Requires a value for ``start_date``.
            platforms: list
                Filter by platforms

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        args = {'link': link,
                'startDate': start_date,
                'endDate': end_date,
                'includeSummary': str(include_summary),
                'platforms': self.list_to_string(platforms)}

        logger.info("Retrieving posts based on link.")
        pt = Table(self.base_request('links', args=args))
        logger.info(f'Retrieved {pt.num_rows} links.')
        self.unpack(pt)
        return pt



import logging
import re
from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)


class Mailchimp():
    """
    Instantiate Mailchimp Class

    `Args:`
        api_key:
            The Mailchimp-provided application key. Not required if
            ``MAILCHIMP_API_KEY`` env variable set.
    `Returns:`
        Mailchimp Class
    """

    def __init__(self, api_key=None):
        self.api_key = check_env.check('MAILCHIMP_API_KEY', api_key)
        self.domain = re.findall("(?<=-).+$", self.api_key)[0]
        self.uri = f'https://{self.domain}.api.mailchimp.com/3.0/'
        self.client = APIConnector(self.uri, auth=('x', self.api_key))

    def get_request(self, endpoint, params=None, **kwargs):
        # Internal method to make a get request.

        r = self.client.request(self.uri + endpoint, 'GET', params=params)
        self.client.validate_response(r)
        data = r.json()

        return data

    def get_lists(self, fields=None, exclude_fields=None,
                  count=None, offset=None, before_date_created=None,
                  since_date_created=None, before_campaign_last_sent=None,
                  since_campaign_last_sent=None, email=None, sort_field=None,
                  sort_dir=None):
        """
        Get a table of lists under the account based on query parameters. Note
        that argument descriptions here are sourced from Mailchimp's official
        API documentation.

        `Args:`
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            before_date_created: string
                Restrict response to lists created before the set date. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            since_date_created: string
                Restrict results to lists created after the set date. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            before_campaign_last_sent: string
                Restrict results to lists created before the last campaign send
                date. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            since_campaign_last_sent: string
                Restrict results to lists created after the last campaign send
                date. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            email: string
                Restrict results to lists that include a specific subscriber's
                email address.
            sort_field: string, can only be 'date_created' or None
                Returns files sorted by the specified field.
            sort_dir: string, can only be 'ASC', 'DESC', or None
                Determines the order direction for sorted results.

        `Returns:`
            Table Class
        """
        params = {'fields': fields,
                  'exclude_fields': exclude_fields,
                  'count': count,
                  'offset': offset,
                  'before_date_created': before_date_created,
                  'since_date_created': since_date_created,
                  'before_campaign_last_sent': before_campaign_last_sent,
                  'since_campaign_last_sent': since_campaign_last_sent,
                  'email': email,
                  'sort_field': sort_field,
                  'sort_dir': sort_dir}

        response = self.get_request('lists', params=params)
        tbl = Table(response['lists'])
        logger.info(f'Found {tbl.num_rows} lists.')
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_campaigns(self, fields=None, exclude_fields=None,
                      count=None, offset=None, type=None, status=None,
                      before_send_time=None, since_send_time=None,
                      before_create_time=None, since_create_time=None,
                      list_id=None, folder_id=None, member_id=None,
                      sort_field=None, sort_dir=None):
        """
        Get a table of campaigns under the account based on query parameters.
        Note that argument descriptions here are sourced from Mailchimp's
        official API documentation.

        `Args:`
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            type: string, can only be 'regular', 'plaintext', 'absplit', 'rss',
            'variate', or None
                The campaign type.
            status: string, can only be 'save', 'paused', 'schedule',
            'sending', 'sent', or None
                The status of the campaign.
            before_send_time: string
                Restrict the response to campaigns sent before the set time. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            since_send_time: string
                Restrict the response to campaigns sent after the set time. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            before_create_time: string
                Restrict the response to campaigns created before the set time.
                We recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            since_create_time: string
                Restrict the response to campaigns created after the set time.
                We recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            list_id: string
                The unique id for the list.
            folder_id: string
                The unique folder id.
            member_id: string
                Retrieve campaigns sent to a particular list member. Member ID
                is The MD5 hash of the lowercase version of the list member’s
                email address.
            sort_field: string, can only be 'create_time', 'send_time', or None
                Returns files sorted by the specified field.
            sort_dir: string, can only be 'ASC', 'DESC', or None
                Determines the order direction for sorted results.

        `Returns:`
            Table Class
        """
        params = {'fields': fields,
                  'exclude_fields': exclude_fields,
                  'count': count,
                  'offset': offset,
                  'type': type,
                  'status': status,
                  'before_send_time': before_send_time,
                  'since_send_time': since_send_time,
                  'before_create_time': before_create_time,
                  'since_create_time': since_create_time,
                  'list_id': list_id,
                  'folder_id': folder_id,
                  'member_id': member_id,
                  'sort_field': sort_field,
                  'sort_dir': sort_dir}

        response = self.get_request('campaigns', params=params)
        tbl = Table(response['campaigns'])
        logger.info(f'Found {tbl.num_rows} campaigns.')
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_members(self, list_id, fields=None,
                    exclude_fields=None, count=None, offset=None,
                    email_type=None, status=None, since_timestamp_opt=None,
                    before_timestamp_opt=None, since_last_changed=None,
                    before_last_changed=None, unique_email_id=None,
                    vip_only=False, interest_category_id=None,
                    interest_ids=None, interest_match=None, sort_field=None,
                    sort_dir=None, since_last_campaign=None,
                    unsubscribed_since=None):
        """
        Get a table of members in a list based on query parameters. Note that
        argument descriptions here are sourced from Mailchimp's official API
        documentation.

        `Args:`
            list_id: string
                The unique ID of the list to fetch members from.
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of fields as strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            email_type: string
                The email type.
            status: string, can only be 'subscribed', 'unsubscribed',
            'cleaned', 'pending', 'transactional', 'archived', or None
                The subscriber's status.
            since_timestamp_opt: string
                Restrict results to subscribers who opted-in after the set
                timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            before_timestamp_opt: string
                Restrict results to subscribers who opted-in before the set
                timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            since_last_changed: string
                Restrict results to subscribers whose information changed after
                the set timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            before_last_changed: string
                Restrict results to subscribers whose information changed
                before the set timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            unique_email_id: string
                A unique identifier for the email address across all Mailchimp
                lists. This parameter can be found in any links with Ecommerce
                Tracking enabled.
            vip_only: boolean
                A filter to return only the list's VIP members. Passing true
                will restrict results to VIP list members, passing false will
                return all list members.
            interest_category_id: string
                The unique id for the interest category.
            interest_ids: list of strings
                Used to filter list members by interests. Must be accompanied
                by interest_category_id and interest_match. The value must be a
                comma separated list of interest ids present for any supplied
                interest categories.
            interest_match: string, can only be 'any', 'all', 'none', or None
                Used to filter list members by interests. Must be accompanied
                by interest_category_id and interest_ids. "any" will match a
                member with any of the interest supplied, "all" will only match
                members with every interest supplied, and "none" will match
                members without any of the interest supplied.
            sort_field: string, can only be 'timestamp_opt',
            'timestamp_signup', 'last_changed', or None
                Returns files sorted by the specified field.
            sort_dir: string, can only be 'ASC', 'DESC', or None
                Determines the order direction for sorted results.
            since_last_campaign: string
                Filter subscribers by those
                subscribed/unsubscribed/pending/cleaned since last email
                campaign send. Member status is required to use this filter.
            unsubscribed_since: string
                Filter subscribers by those unsubscribed since a specific date.
                Using any status other than unsubscribed with this filter will
                result in an error.

        `Returns:`
            Table Class
        """
        params = {'fields': fields,
                  'exclude_fields': exclude_fields,
                  'count': count,
                  'offset': offset,
                  'email_type': email_type,
                  'status': status,
                  'since_timestamp_opt': since_timestamp_opt,
                  'before_timestamp_opt': before_timestamp_opt,
                  'since_last_changed': since_last_changed,
                  'before_last_changed': before_last_changed,
                  'unqiue_email_id': unique_email_id,
                  'vip_only': vip_only,
                  'interest_category_id': interest_category_id,
                  'interest_ids': interest_ids,
                  'interest_match': interest_match,
                  'sort_field': sort_field,
                  'sort_dir': sort_dir,
                  'since_last_campaign': since_last_campaign,
                  'unsubscribed_since': unsubscribed_since}

        response = self.get_request(f'lists/{list_id}/members', params=params)
        tbl = Table(response['members'])
        logger.info(f'Found {tbl.num_rows} members.')
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_campaign_emails(self, campaign_id, fields=None,
                            exclude_fields=None, count=None, offset=None,
                            since=None):
        """
        Get a table of individual emails from a campaign based on query
        parameters. Note that argument descriptions here are sourced from
        Mailchimp's official API documentation.

        `Args:`
            campaign_id: string
                The unique ID of the campaign to fetch emails from.
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            since: string
                Restrict results to email activity events that occur after a
                specific time. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.

        `Returns:`
            Table Class
        """
        params = {'fields': fields,
                  'exclude_fields': exclude_fields,
                  'count': count,
                  'offset': offset,
                  'since': since}

        response = self.get_request(f'reports/{campaign_id}/email-activity',
                                    params=params)
        tbl = Table(response['emails'])
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_unsubscribes(self, campaign_id, fields=None,
                         exclude_fields=None, count=None, offset=None):
        """
        Get a table of unsubscribes associated with a campaign based on query
        parameters. Note that argument descriptions here are sourced from
        Mailchimp's official API documentation.

        `Args:`
            campaign_id: string
                The unique ID of the campaign to fetch unsubscribes from.
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.

        `Returns:`
            Table Class
        """
        params = {'fields': fields,
                  'exclude_fields': exclude_fields,
                  'count': count,
                  'offset': offset}

        response = self.get_request(f'reports/{campaign_id}/unsubscribed',
                                    params=params)
        tbl = Table(response['unsubscribes'])
        logger.info(f'Found {tbl.num_rows} unsubscribes for {campaign_id}.')
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()



import unittest
import os
import requests_mock
import unittest.mock as mock
from parsons.ngpvan.van import VAN
from parsons.etl.table import Table
from test.utils import validate_list, assert_matching_tables
from parsons.utilities import cloud_storage


class TestSavedLists(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_saved_lists(self, m):

        json = {'count': 1, 'items': [
            {"savedListId": 517612,
             "listCount": 974656,
             "name": "LikelyParents(16andunder)_DWID_S... - MN",
             "doorCount": 520709,
             "description": "null"
             }
        ], 'nextPageLink': None}

        m.get(self.van.connection.uri + 'savedLists', json=json)

        expected = ['savedListId', 'listCount', 'name', 'doorCount','description']

        self.assertTrue(validate_list(expected, self.van.get_saved_lists()))

    @requests_mock.Mocker()
    def test_get_saved_list(self, m):

        saved_list_id = 517612

        json = {"savedListId": 517612,
                "listCount": 974656,
                "name": "LikelyParents(16andunder)_DWID_S... - MN",
                "doorCount": 520709,
                "description": "null"
                }

        m.get(self.van.connection.uri + f'savedLists/{saved_list_id}', json=json)

        expected = ['savedListId', 'listCount', 'name', 'doorCount', 'description']

        self.assertEqual(self.van.get_saved_list(saved_list_id), json)

    def test_upload_saved_list(self):

        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = 'https://box.com/my_file.zip'

        self.van.connection._soap_client = mock.MagicMock()
        self.van.get_folders = mock.MagicMock()
        self.van.get_folders.return_value = [{'folderId': 1}]

        tbl = Table([['VANID'],['1'],['2'],['3']])
        self.van.upload_saved_list(tbl, 'GOTV List', 1, replace=True, url_type='S3', bucket='tmc-scratch')
        assert self.van.connection._soap_client.service.CreateAndStoreSavedList.called

    @requests_mock.Mocker()
    def test_get_folders(self, m):

        json = {u'count': 2,
                u'items': [
                    {
                        u'folderId': 5046,
                        u'name': u'#2018_MN_active_universe'
                    },
                    {u'folderId': 2168,
                     u'name': u'API Generated Lists'
                     }
                ], u'nextPageLink': None}

        m.get(self.van.connection.uri + 'folders', json=json)

        expected = ['folderId', 'name']

        self.assertTrue(validate_list(expected, self.van.get_folders()))

    @requests_mock.Mocker()
    def test_get_folder(self, m):

        folder_id = 5046

        json = {"folderId": 5046, "name": "#2018_MN_active_universe"}

        m.get(self.van.connection.uri + f'folders/{folder_id}', json=json)

        self.assertEqual(json, self.van.get_folder(folder_id))

    @requests_mock.Mocker()
    def test_export_job_types(self, m):

        json = {u'count': 1, u'items':
                [{u'exportJobTypeId': 4, u'name': u'SavedListExport'}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'exportJobTypes', json=json)

        expected = ['exportJobTypeId', 'name']

        self.assertTrue(validate_list(expected, self.van.get_export_job_types()))

    @requests_mock.Mocker()
    def test_export_job_create(self, m):

        saved_list_id = 517612

        json = {"status": "Completed",
                "errorCode": "null",
                "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "activistCodes": "null",
                "canvassFileRequestId": 448,
                "dateExpired": "2018-09-08T16:04:00Z",
                "surveyQuestions": "null",
                "webhookUrl": "https://www.nothing.com/",
                "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",  # noqa: E501
                "savedListId": 517612,
                "districtFields": "null",
                "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "customFields": "null",
                "type": 4,
                "exportJobId": 448}

        m.post(self.van.connection.uri + 'exportJobs', json=json, status_code=201)

        expected = [
            'status',
            'errorCode',
            'exportJobGuid',
            'activistCodes',
            'canvassFileRequestId',
            'dateExpired',
            'surveyQuestions',
            'webhookUrl',
            'downloadUrl',
            'savedListId',
            'districtFields',
            'canvassFileRequestGuid',
            'customFields',
            'type',
            'exportJobId']

        self.assertEqual(json,self.van.export_job_create(saved_list_id))

    @requests_mock.Mocker()
    def test_get_export_job(self, m):

        export_job_id = 448

        json = {"status": "Completed",
                "errorCode": "null",
                "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "activistCodes": "null",
                "canvassFileRequestId": 448,
                "dateExpired": "2018-09-08T16:04:00Z",
                "surveyQuestions": "null",
                "webhookUrl": "https://www.nothing.com/",
                "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",  # noqa: E501
                "savedListId": 517612,
                "districtFields": "null",
                "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
                "customFields": "null",
                "type": 4,
                "exportJobId": 448}

        expected = [
            'status',
            'errorCode',
            'exportJobGuid',
            'activistCodes',
            'canvassFileRequestId',
            'dateExpired',
            'surveyQuestions',
            'webhookUrl',
            'downloadUrl',
            'savedListId',
            'districtFields',
            'canvassFileRequestGuid',
            'customFields',
            'type',
            'exportJobId']

        m.get(self.van.connection.uri + f'exportJobs/{export_job_id}', json=json)

        self.assertEqual(json, self.van.get_export_job(export_job_id))


from requests import request as _request
from parsons.etl.table import Table
from parsons.utilities.date_convert import iso_to_unix
import petl
import re
import os
import logging

logger = logging.getLogger(__name__)

MA_URI = 'http://events.mobilizeamerica.io/api/v1/'


class MobilizeAmerica(object):
    """
    Instantiate MobilizeAmerica Class

    api_key: str
        An api key issued by Mobilize America. This is required to access some private methods.

    `Returns:`
        MobilizeAmerica Class
    """

    def __init__(self, api_key=None):

        self.uri = MA_URI
        self.api_key = api_key or os.environ.get('MOBILIZE_AMERICA_API_KEY')

        if not self.api_key:
            logger.info('Mobilize America API Key missing. Calling methods that rely on private'
                        ' endpoints will fail.')

    def request(self, url, req_type='GET', post_data=None, args=None, auth=False):
        # Internal request method

        if auth:

            if not self.api_key:
                raise TypeError('This method requires an api key.')
            else:
                header = {'Authorization': 'Bearer ' + self.api_key}

        else:
            header = None

        r = _request(req_type, url, json=post_data, params=args, headers=header)

        if 'error' in r.json():
            raise ValueError('API Error:' + str(r.json()['error']))

        return r

    def request_paginate(self, url, req_type='GET', post_data=None, args=None, raw=False,
                         paginate=False, auth=False):

        r = self.request(url, req_type=req_type, args=args, auth=auth)

        json = r.json()['data']

        while r.json()['next']:

            r = self.request(r.json()['next'], req_type=req_type)
            json.extend(r.json()['data'])

        return json

    def _time_parse(self, time_arg):
        # Parse the date filters

        trans = [('>=', 'gte_'),
                 ('>', 'gt_'),
                 ('<=', 'lte_'),
                 ('<', 'lt_')]

        if time_arg:

            time = re.sub('<=|<|>=|>', '', time_arg)
            time = iso_to_unix(time)
            time_filter = re.search('<=|<|>=|>', time_arg).group()

            for i in trans:
                if time_filter == i[0]:
                    return i[1] + str(time)

            raise ValueError('Invalid time operator. Must be one of >=, >, <= or >.')

        return time_arg

    def get_organizations(self, updated_since=None):
        """
        Return all active organizations on the platform.

        `Args:`
            updated_since: str
                Filter to organizations updated since given date (ISO Date)
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        return Table(self.request_paginate(self.uri + 'organizations',
                                           args={'updated_since': iso_to_unix(updated_since)}))

    def get_events(self, organization_id=None, updated_since=None, timeslot_start=None,
                   timeslot_end=None, timeslots_table=False, max_timeslots=None):
        """
        Fetch all public events on the platform.

        `Args:`
            organization_id: list or int
                Filter events by a single or multiple organization ids
            updated_since: str
                Filter to events updated since given date (ISO Date)
            timeslot_start: str
                Filter by a timeslot start of events using ``>``,``>=``,``<``,``<=``
                operators and ISO date (ex. ``<=2018-12-13 05:00:00PM``)
            timeslot_end: str
                Filter by a timeslot end of events using ``>``,``>=``,``<``,``<=``
                operators and ISO date (ex. ``<=2018-12-13 05:00:00PM``)
            timeslot_table: boolean
                Return timeslots as a separate long table. Useful for extracting
                to databases.
            max_timeslots: int
                If not returning a timeslot table, will unpack time slots. If do not
                set this kwarg, it will add a column for each time slot. The argument
                limits the number of columns and discards any additional timeslots
                after that.

                For example: If there are 20 timeslots associated with your event,
                and you set the max time slots to 5, it will only return the first 5
                time slots as ``time_slot_0``, ``time_slot_1`` etc.

                This is helpful in situations where you have a regular sync
                running and want to ensure that the column headers remain static.

        `Returns`
            Parsons Table or dict or Parsons Tables
                See :ref:`parsons-table` for output options.
        """

        if isinstance(organization_id, (str, int)):
            organization_id = [organization_id]

        args = {'organization_id': organization_id,
                'updated_since': iso_to_unix(updated_since),
                'timeslot_start': self._time_parse(timeslot_start),
                'timeslot_end': self._time_parse(timeslot_end)}

        tbl = Table(self.request_paginate(self.uri + 'events', args=args))

        if tbl.num_rows > 0:

            tbl.unpack_dict('sponsor')
            tbl.unpack_dict('location', prepend=False)
            tbl.unpack_dict('location', prepend=False)  # Intentional duplicate
            tbl.table = petl.convert(tbl.table, 'address_lines', lambda v: ' '.join(v))

            if timeslots_table:

                timeslots_tbl = tbl.long_table(['id'], 'timeslots', 'event_id')
                return {'events': tbl, 'timeslots': timeslots_tbl}

            else:
                tbl.unpack_list('timeslots', replace=True, max_columns=max_timeslots)
                cols = tbl.columns
                for c in cols:
                    if re.search('timeslots', c, re.IGNORECASE) is not None:
                        tbl.unpack_dict(c)

        return tbl

    def get_events_organization(self, organization_id=None, updated_since=None, timeslot_start=None,
                                timeslot_end=None, timeslots_table=False, max_timeslots=None):
        """
        Fetch all public events for an organization. This includes both events owned
        by the organization (as indicated by the organization field on the event object)
        and events of other organizations promoted by this specified organization.

        .. note::
            API Key Required

        `Args:`
            organization_id: list or int
                Filter events by a single or multiple organization ids
            updated_since: str
                Filter to events updated since given date (ISO Date)
            timeslot_start: str
                Filter by a timeslot start of events using ``>``,``>=``,``<``,``<=``
                operators and ISO date (ex. ``<=2018-12-13 05:00:00PM``)
            timeslot_end: str
                Filter by a timeslot end of events using ``>``,``>=``,``<``,``<=``
                operators and ISO date (ex. ``<=2018-12-13 05:00:00PM``)
            timeslot_table: boolean
                Return timeslots as a separate long table. Useful for extracting
                to databases.
            zipcode: str
                Filter by a Events' Locations' postal code. If present, returns Events
                sorted by distance from zipcode. If present, virtual events will not be returned.
            max_dist: str
                Filter Events' Locations' distance from provided zipcode.
            visibility: str
                Either `PUBLIC` or `PRIVATE`. Private events only return if user is authenticated;
                if `visibility=PRIVATE` and user doesn't have permission, no events returned.
            exclude_full: bool
                If `exclude_full=true`, filter out full Timeslots (and Events if all of an Event's
                Timeslots are full)
            is_virtual: bool
                `is_virtual=false` will return only in-person events, while `is_virtual=true` will
                return only virtual events. If excluded, return virtual and in-person events. Note
                that providing a zipcode also implies `is_virtual=false`.
            event_types:enum
                The type of the event, one of: `CANVASS`, `PHONE_BANK`, `TEXT_BANK`, `MEETING`,
                `COMMUNITY`, `FUNDRAISER`, `MEET_GREET`, `HOUSE_PARTY`, `VOTER_REG`, `TRAINING`,
                `FRIEND_TO_FRIEND_OUTREACH`, `DEBATE_WATCH_PARTY`, `ADVOCACY_CALL`, `OTHER`.
                This list may expand in the future.
            max_timeslots: int
                If not returning a timeslot table, will unpack time slots. If do not
                set this arg, it will add a column for each time slot. The argument
                limits the number of columns and discards any additional timeslots
                after that.

                For example: If there are 20 timeslots associated with your event,
                and you set the max time slots to 5, it will only return the first 5
                time slots as ``time_slot_0``, ``time_slot_1`` etc.

                This is helpful in situations where you have a regular sync
                running and want to ensure that the column headers remain static.

        `Returns`
            Parsons Table or dict or Parsons Tables
                See :ref:`parsons-table` for output options.
        """

        if isinstance(organization_id, (str, int)):
            organization_id = [organization_id]

        args = {'organization_id': organization_id,
                'updated_since': iso_to_unix(updated_since),
                'timeslot_start': self._time_parse(timeslot_start),
                'timeslot_end': self._time_parse(timeslot_end),
                }

        tbl = Table(self.request_paginate(self.uri + 'events', args=args, auth=True))

        if tbl.num_rows > 0:

            tbl.unpack_dict('sponsor')
            tbl.unpack_dict('location', prepend=False)
            tbl.unpack_dict('location', prepend=False)  # Intentional duplicate
            tbl.table = petl.convert(tbl.table, 'address_lines', lambda v: ' '.join(v))

            if timeslots_table:

                timeslots_tbl = tbl.long_table(['id'], 'timeslots', 'event_id')
                return {'events': tbl, 'timeslots': timeslots_tbl}

            else:
                tbl.unpack_list('timeslots', replace=True, max_columns=max_timeslots)
                cols = tbl.columns
                for c in cols:
                    if re.search('timeslots', c, re.IGNORECASE) is not None:
                        tbl.unpack_dict(c)

        return tbl

    def get_events_deleted(self, organization_id=None, updated_since=None):
        """
        Fetch deleted public events on the platform.

        `Args:`
            organization_id: list or int
                Filter events by a single or multiple organization ids
            updated_since: str
                Filter to events updated since given date (ISO Date)
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if isinstance(organization_id, (str, int)):
            organization_id = [organization_id]

        args = {'organization_id': organization_id,
                'updated_since': iso_to_unix(updated_since)}

        return Table(self.request_paginate(self.uri + 'events/deleted', args=args))

    def get_people(self, organization_id=None, updated_since=None):
        """
        Fetch all people (volunteers) who are affiliated with the organization.

        .. note::
            API Key Required

        `Args:`
            organization_id: list of int
                Filter events by a single or multiple organization ids
            updated_since: str
                Filter to events updated since given date (ISO Date)
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.uri + 'organizations/' + str(organization_id) + '/people'

        return Table(self.request_paginate(url,
                                           args={'updated_since': iso_to_unix(updated_since)},
                                           auth=True))

    def get_attendances(self, organization_id=None, updated_since=None):
        """
        Fetch all attendances which were either promoted by the organization or
        were for events owned by the organization.

        .. note::
            API Key Required

        `Args:`
            organization_id: list of int
                Filter events by a single or multiple organization ids
            updated_since: str
                Filter to events updated since given date (ISO Date)
        `Returns`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.uri + 'organizations/' + str(organization_id) + '/attendances'

        return Table(self.request_paginate(url,
                                           args={'updated_since': iso_to_unix(updated_since)},
                                           auth=True))



import json
import os
import unittest
from unittest import mock
from parsons.action_kit.action_kit import ActionKit
from parsons.etl.table import Table

from test.utils import assert_matching_tables

ENV_PARAMETERS = {
    'ACTION_KIT_DOMAIN': 'env_domain',
    'ACTION_KIT_USERNAME': 'env_username',
    'ACTION_KIT_PASSWORD': 'env_password'
}


class TestActionKit(unittest.TestCase):

    def setUp(self):
        self.actionkit = ActionKit(
            domain='domain.actionkit.com',
            username='user',
            password='password'
        )
        self.actionkit.conn = mock.MagicMock()

    def tearDown(self):
        pass

    @mock.patch.dict(os.environ, ENV_PARAMETERS)
    def test_from_envrion(self):
        actionkit = ActionKit()
        self.assertEqual(actionkit.domain, 'env_domain')
        self.assertEqual(actionkit.username, 'env_username')
        self.assertEqual(actionkit.password, 'env_password')

    def test_base_endpoint(self):
        # Test the endpoint
        url = self.actionkit._base_endpoint('user')
        self.assertEqual(url, 'https://domain.actionkit.com/rest/v1/user/')

        url = self.actionkit._base_endpoint('user', 1234)
        self.assertEqual(url, 'https://domain.actionkit.com/rest/v1/user/1234/')

        url = self.actionkit._base_endpoint('user', '1234')
        self.assertEqual(url, 'https://domain.actionkit.com/rest/v1/user/1234/')

    def test_get_user(self):
        # Test get user
        self.actionkit.get_user(123)
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/user/123/',
        )

    def test_get_user_fields(self):
        self.actionkit.get_user_fields()
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/user/schema/'
        )

    def test_create_user(self):
        # Test create user

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_user(email='test')
        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/user/',
            data=json.dumps({'email': 'test'})
        )

    def test_update_user(self):
        # Test update user

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=202)
        self.actionkit.conn = resp_mock

        self.actionkit.update_user(123, last_name='new name')
        self.actionkit.conn.patch.assert_called_with(
            'https://domain.actionkit.com/rest/v1/user/123/',
            data=json.dumps({'last_name': 'new name'})
        )

    def test_delete_user(self):
        # Test delete user

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.patch()).status_code = mock.PropertyMock(return_value=204)
        self.actionkit.conn = resp_mock

        self.actionkit.delete_user(123)
        self.actionkit.conn.delete.assert_called_with(
            'https://domain.actionkit.com/rest/v1/user/123/',
        )

    def test_get_campaign(self):
        # Test get campaign
        self.actionkit.get_campaign(123)
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/campaign/123/',
        )

    def test_create_campaign(self):
        # Test create campaign

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_campaign(name='new_campaign', field='field')
        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/campaign/',
            data=json.dumps({
                'name': 'new_campaign',
                'field': 'field'
            })
        )

    def test_get_event_create_page(self):
        # Test get event create page
        self.actionkit.get_event_create_page(123)
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventcreatepage/123/',
        )

    def test_create_event_create_page(self):
        # Test create event create page

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_create_page(
            name='new_page',
            campaign_id='123',
            title='title'
        )
        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventcreatepage/',
            data=json.dumps({
                'campaign': '/rest/v1/campaign/123/',
                'name': 'new_page',
                'title': 'title'
            })
        )

    def test_get_event_create_form(self):
        # Test get event create form
        self.actionkit.get_event_create_form(123)
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventcreateform/123/',
        )

    def test_create_event_create_form(self):
        # Test event create form

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_create_form(
            page_id='123',
            thank_you_text='thank you'
        )
        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventcreateform/',
            data=json.dumps({
                'page': '/rest/v1/eventcreatepage/123/',
                'thank_you_text': 'thank you'
            })
        )

    def test_get_event_signup_page(self):
        # Test get event signup page
        self.actionkit.get_event_signup_page(123)
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventsignuppage/123/',
        )

    def test_create_event_signup_page(self):
        # Test create event signup page

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_signup_page(
            name='new_name',
            campaign_id='123',
            title='title'
        )
        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventsignuppage/',
            data=json.dumps({
                'campaign': '/rest/v1/campaign/123/',
                'name': 'new_name',
                'title': 'title'
            })
        )

    def test_get_event_signup_form(self):
        # Test get event signup form
        self.actionkit.get_event_signup_form(123)
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventsignupform/123/',
        )

    def test_create_event_signup_form(self):
        # Test create event signup form

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_event_signup_form(
            page_id='123',
            thank_you_text='thank you'
        )
        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/eventsignupform/',
            data=json.dumps({
                'page': '/rest/v1/page/123/',
                'thank_you_text': 'thank you'
            })
        )

    def test_get_page_followup(self):
        # Test get page followup
        self.actionkit.get_page_followup(123)
        self.actionkit.conn.get.assert_called_with(
            'https://domain.actionkit.com/rest/v1/pagefollowup/123/',
        )

    def test_create_page_followup(self):
        # Test create page followup

        # Mock resp and status code
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_page_followup(
            signup_page_id='123',
            url='url'
        )
        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/pagefollowup/',
            data=json.dumps({
                'page': '/rest/v1/eventsignuppage/123/',
                'url': 'url'
            })
        )

    def test_create_generic_action(self):
        # Test create a generic action

        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit.conn = resp_mock

        self.actionkit.create_generic_action(email='bob@bob.com', page='my_action')

        self.actionkit.conn.post.assert_called_with(
            'https://domain.actionkit.com/rest/v1/action/',
            data=json.dumps({'email': 'bob@bob.com', 'page': 'my_action'}))

    def test_bulk_upload_table(self):
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit._conn = lambda self: resp_mock
        self.actionkit.bulk_upload_table(
            Table([('user_id', 'user_customfield1', 'action_foo'), (5, 'yes', '123 Main St')]),
            'fake_page')
        self.assertEqual(resp_mock.post.call_count, 2)
        name, args, kwargs = resp_mock.method_calls[1]
        self.assertEqual(kwargs['data'],
                         {'page': 'fake_page', 'autocreate_user_fields': 0, 'user_fields_only': 0})
        upload_data = kwargs['files']['upload'].read()
        self.assertEqual(upload_data.decode(),
                         'user_id,user_customfield1,action_foo\r\n5,yes,123 Main St\r\n')

    def test_bulk_upload_table_userfields(self):
        resp_mock = mock.MagicMock()
        type(resp_mock.post()).status_code = mock.PropertyMock(return_value=201)
        self.actionkit._conn = lambda self: resp_mock
        self.actionkit.bulk_upload_table(
            Table([('user_id', 'user_customfield1'), (5, 'yes')]),
            'fake_page')
        self.assertEqual(resp_mock.post.call_count, 2)
        name, args, kwargs = resp_mock.method_calls[1]
        self.assertEqual(kwargs['data'],
                         {'page': 'fake_page', 'autocreate_user_fields': 0, 'user_fields_only': 1})
        self.assertEqual(kwargs['files']['upload'].read().decode(),
                         'user_id,user_customfield1\r\n5,yes\r\n')

    def test_table_split(self):
        test1 = Table([('x', 'y', 'z'), ('a', 'b', ''), ('1', '', '3'), ('4', '', '6')])
        tables = self.actionkit._split_tables_no_empties(test1)
        self.assertEqual(len(tables), 2)
        assert_matching_tables(tables[0], Table([('x', 'y'), ('a', 'b')]))
        assert_matching_tables(tables[1], Table([('x', 'z'), ('1', '3'), ('4', '6')]))

        test2 = Table([('x', 'y', 'z'), ('a', 'b', 'c'), ('1', '2', '3'), ('4', '5', '6')])
        tables2 = self.actionkit._split_tables_no_empties(test2)
        self.assertEqual(len(tables2), 1)
        assert_matching_tables(tables2[0], test2)



from parsons import Table
from parsons.utilities import check_env
import petl
import mysql.connector as mysql
from contextlib import contextmanager
from parsons.utilities import files
import pickle
import logging
import os
from parsons.databases.table import BaseTable
from parsons.databases.mysql.create_table import MySQLCreateTable

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000

logger = logging.getLogger(__name__)


class MySQL(MySQLCreateTable):
    """
    Connect to a MySQL database.

    `Args:`
        username: str
            Required if env variable ``MYSQL_USERNAME`` not populated
        password: str
            Required if env variable ``MYSQL_PASSWORD`` not populated
        host: str
            Required if env variable ``MYSQL_HOST`` not populated
        db: str
            Required if env variable ``MYSQL_DB`` not populated
        port: int
            Can be set by env variable ``MYSQL_PORT`` or argument.
    """

    def __init__(self, host=None, username=None, password=None, db=None, port=3306):

        self.username = check_env.check('MYSQL_USERNAME', username)
        self.password = check_env.check('MYSQL_PASSWORD', password)
        self.host = check_env.check('MYSQL_HOST', host)
        self.db = check_env.check('MYSQL_DB', db)
        self.port = port or os.environ.get('MYSQL_PORT')

    @contextmanager
    def connection(self):
        """
        Generate a MySQL connection. The connection is set up as a python "context manager", so
        it will be closed automatically (and all queries committed) when the connection goes out
        of scope.

        When using the connection, make sure to put it in a ``with`` block (necessary for
        any context manager):
        ``with mysql.connection() as conn:``

        `Returns:`
            MySQL `connection` object
        """

        # Create a mysql connection and cursor
        connection = mysql.connect(host=self.host,
                                   user=self.username,
                                   passwd=self.password,
                                   database=self.db,
                                   port=self.port)

        try:
            yield connection
        except mysql.Error:
            connection.rollback()
            raise
        else:
            connection.commit()
        finally:
            connection.close()

    @contextmanager
    def cursor(self, connection):
        cur = connection.cursor(buffered=True)

        try:
            yield cur
        finally:
            cur.close()

    def query(self, sql, parameters=None):
        """
        Execute a query against the database. Will return ``None``if the query returns zero rows.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the `mysql style <https://security.openstack.org/guidelines/dg_parameterize-database-queries.html>`_.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

            # Note that the name contains a quote, which could break your query if not escaped
            # properly.
            name = "Beatrice O'Brady"
            sql = "SELECT * FROM my_table WHERE name = %s"
            mysql.query(sql, parameters=[name])

        .. code-block:: python

            names = ["Allen Smith", "Beatrice O'Brady", "Cathy Thompson"]
            placeholders = ', '.join('%s' for item in names)
            sql = f"SELECT * FROM my_table WHERE name IN ({placeholders})"
            mysql.query(sql, parameters=names)

        `Args:`
            sql: str
                A valid SQL statement
            parameters: list
                A list of python variables to be converted into SQL values in your query

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.

        """  # noqa: E501

        with self.connection() as connection:
            return self.query_with_connection(sql, connection, parameters=parameters)

    def query_with_connection(self, sql, connection, parameters=None, commit=True):
        """
        Execute a query against the database, with an existing connection. Useful for batching
        queries together. Will return ``None`` if the query returns zero rows.

        `Args:`
            sql: str
                A valid SQL statement
            connection: obj
                A connection object obtained from ``mysql.connection()``
            parameters: list
                A list of python variables to be converted into SQL values in your query
            commit: boolean
                Whether to commit the transaction immediately. If ``False`` the transaction will
                be committed when the connection goes out of scope and is closed (or you can
                commit manually with ``connection.commit()``).

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        with self.cursor(connection) as cursor:

            # The python connector can only execute a single sql statement, so we will
            # break up each statement and execute them separately.
            for s in sql.strip().split(';'):
                if len(s) != 0:
                    logger.debug(f'SQL Query: {sql}')
                    cursor.execute(s, parameters)

            if commit:
                connection.commit()

            # If the SQL query provides no response, then return None
            if not cursor.description:
                logger.debug('Query returned 0 rows')
                return None

            else:
                # Fetch the data in batches, and "pickle" the rows to a temp file.
                # (We pickle rather than writing to, say, a CSV, so that we maintain
                # all the type information for each field.)
                temp_file = files.create_temp_file()

                with open(temp_file, 'wb') as f:
                    # Grab the header
                    pickle.dump(cursor.column_names, f)

                    while True:
                        batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                        if len(batch) == 0:
                            break

                        logger.debug(f'Fetched {len(batch)} rows.')
                        for row in batch:
                            pickle.dump(row, f)

                # Load a Table from the file
                final_tbl = Table(petl.frompickle(temp_file))

                logger.debug(f'Query returned {final_tbl.num_rows} rows.')
                return final_tbl

    def copy(self, tbl, table_name, if_exists='fail', chunk_size=1000):
        """
        Copy a :ref:`parsons-table` to the database.

        .. note::
            This method utilizes extended inserts rather `LOAD DATA INFILE` since
            many MySQL Database configurations do not allow data files to be
            loaded. It results in a minor performance hit compared to `LOAD DATA`.

        tbl: parsons.Table
            A Parsons table object
        table_name: str
            The destination schema and table (e.g. ``my_schema.my_table``)
        if_exists: str
            If the table already exists, either ``fail``, ``append``, ``drop``
            or ``truncate`` the table.
        chunk_size: int
            The number of rows to insert per query.
        """

        if tbl.num_rows == 0:
            logger.info('Parsons table is empty. Table will not be created.')
            return None

        with self.connection() as connection:

            # Create table if not exists
            if self._create_table_precheck(connection, table_name, if_exists):
                sql = self.create_statement(tbl, table_name)
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f'Table {table_name} created.')

            # Chunk tables in batches of 1K rows, though this can be tuned and
            # optimized further.
            for t in tbl.chunk(chunk_size):
                sql = self._insert_statement(t, table_name)
                self.query_with_connection(sql, connection, commit=False)

    def _insert_statement(self, tbl, table_name):
        """
        Convert the table data into a string for bulk importing.
        """

        # Single column tables
        if len(tbl.columns) == 1:
            values = [f"({row[0]})" for row in tbl.data]

        # Multi-column tables
        else:
            values = [str(row) for row in tbl.data]

        # Create full insert statement
        sql = f"""INSERT INTO {table_name}
                  ({','.join(tbl.columns)})
                  VALUES {",".join(values)};"""

        return sql

    def _create_table_precheck(self, connection, table_name, if_exists):
        """
        Helper to determine what to do when you need a table that may already exist.

        `Args:`
            connection: obj
                A connection object obtained from ``mysql.connection()``
            table_name: str
                The table to check
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``,
                or ``truncate`` the table.
        `Returns:`
            bool
                True if the table needs to be created, False otherwise.
        """

        if if_exists not in ['fail', 'truncate', 'append', 'drop']:
            raise ValueError("Invalid value for `if_exists` argument")

        # If the table exists, evaluate the if_exists argument for next steps.
        if self.table_exists(table_name):

            if if_exists == 'fail':
                raise ValueError('Table already exists.')

            if if_exists == 'truncate':
                sql = f"TRUNCATE TABLE {table_name}"
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f"{table_name} truncated.")
                return False

            if if_exists == 'drop':
                sql = f"DROP TABLE {table_name}"
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f"{table_name} dropped.")
                return True

        else:
            return True

    def table_exists(self, table_name):
        """
        Check if a table or view exists in the database.

        `Args:`
            table_name: str
                The table name
            view: boolean
                Check to see if a view exists by the same name

        `Returns:`
            boolean
                ``True`` if the table exists and ``False`` if it does not.
        """

        if self.query(f"SHOW TABLES LIKE '{table_name}'").first == table_name:
            return True
        else:
            return False

    def table(self, table_name):
        # Return a MySQL table object

        return MySQLTable(self, table_name)


class MySQLTable(BaseTable):
    # MySQL table object.

    def get_rows(self, offset=0, chunk_size=None):
        """
        Get rows from a table.
        """

        sql = f"SELECT * FROM {self.table}"

        if chunk_size:
            sql += f" LIMIT {chunk_size}"

        if offset:
            sql += f" ,{offset}"

        return self.db.query(sql)

    def get_new_rows(self, primary_key, cutoff_value, offset=0, chunk_size=None):
        """
        Get rows that have a greater primary key value than the one
        provided.

        It will select every value greater than the provided value.
        """

        sql = f"""
               SELECT
               *
               FROM {self.table}
               WHERE {primary_key} > {cutoff_value}
               """

        if chunk_size:
            sql += f" LIMIT {chunk_size}, {offset};"

        return self.db.query(sql)



import smtplib

from parsons.notifications.sendmail import SendMail
from parsons.utilities.check_env import check


class SMTP(SendMail):
    """Create a SMTP object, for sending emails.

    `Args:`
        host: str
            The host of the SMTP server
        port: int
            The port of the SMTP server (Default is 587 for TLS)
        username: str
            The username of the SMTP server login
        password: str
            The password of the SMTP server login
        tls: bool
            Defaults to True -- pass "0" or "False" to SMTP_TLS to disable
        close_manually: bool
            When set to True, send_message will not close the connection
    """
    def __init__(self, host=None, port=None, username=None, password=None, tls=None,
                 close_manually=False):
        self.host = check('SMTP_HOST', host)
        self.port = check('SMTP_PORT', port, optional=True) or 587
        self.username = check('SMTP_USER', username)
        self.password = check('SMTP_PASSWORD', password)
        self.tls = not (check('SMTP_TLS', tls, optional=True) in ('false', 'False', '0', False))
        self.close_manually = close_manually

        self.conn = None

    def get_connection(self):
        if self.conn is None:
            self.conn = smtplib.SMTP(self.host, self.port)
            self.conn.ehlo()
            if self.tls:
                self.conn.starttls()
            self.conn.login(self.username, self.password)
        return self.conn

    def _send_message(self, message):
        """Send an email message.

        `Args:`
            message: `MIME object <https://docs.python.org/2/library/email.mime.html>`
                i.e. the objects created by the create_* instance methods
        `Returns:`
            dict of refused To addresses (otherwise None)
        """
        self.log.info("Sending a message...")
        try:
            conn = self.get_connection()
            result = conn.sendmail(message['From'],
                                   [x.strip() for x in message['To'].split(',')],
                                   message.as_string())
        except Exception:
            self.log.exception(
                'An error occurred: while attempting to send a message.')
            raise

        if result:
            self.log.warning("Message failed to send to some recipients: " + str(result))
        if not self.close_manually:
            conn.quit()
            self.conn = None
        return result



import unittest
import os
import requests_mock
from parsons.ngpvan.van import VAN
from test.utils import validate_list


os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestNGPVAN(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters")

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_events(self, m):

        json = {'count': 6, 'items': [
            {"eventId": 1062,
             "startDate": "2010-05-25T11:00:00-05:00",
             "codes": "null",
             "endDate": "2010-05-25T15:00:00-05:00",
             "name": "Sample",
             "roles": "null",
             "isOnlyEditableByCreatingUser": "true",
             "ticketCategories": "null",
             "eventType": {
                                 "eventTypeId": 29166,
                                 "name": "Confirmation Calls"
             },
             "notes": "null",
             "districtFieldValue": "null",
             "locations": "null",
             "shifts": "null",
             "voterRegistrationBatches": "null",
             "createdDate": "2010-05-25T11:55:00Z",
             "financialProgram": "null",
             "shortName": "Sample",
             "isPubliclyViewable": "null",
             "isActive": "true",
             "description": "This is a sample"
             }],
            'nextPageLink': None}

        m.get(self.van.connection.uri + 'events', json=json)

        # Expected Structure
        expected = [
            'eventId',
            'startDate',
            'codes',
            'endDate',
            'name',
            'roles',
            'isOnlyEditableByCreatingUser',
            'ticketCategories',
            'eventType',
            'notes',
            'districtFieldValue',
            'locations',
            'shifts',
            'voterRegistrationBatches',
            'createdDate',
            'financialProgram',
            'shortName',
            'isPubliclyViewable',
            'isActive',
            'description']

        self.assertTrue(validate_list(expected, self.van.get_events()))

    @requests_mock.Mocker()
    def test_get_event(self, m):

        event_id = 1062

        json = {"eventId": 1062,
                "startDate": "2010-05-25T11:00:00-05:00",
                "codes": "null",
                "endDate": "2010-05-25T15:00:00-05:00",
                "name": "Sample",
                "roles": "null",
                "isOnlyEditableByCreatingUser": "true",
                "ticketCategories": "null",
                "eventType": {
                         "eventTypeId": 29166,
                         "name": "Confirmation Calls"
                },
                "notes": "null",
                "districtFieldValue": "null",
                "locations": "null",
                "shifts": "null",
                "voterRegistrationBatches": "null",
                "createdDate": "2010-05-25T11:55:00Z",
                "financialProgram": "null",
                "shortName": "Sample",
                "isPubliclyViewable": "null",
                "isActive": "true",
                "description": "This is a sample"
                }

        m.get(self.van.connection.uri + 'events/{}'.format(event_id), json=json)

        self.assertEqual(json, self.van.get_event(event_id))

    @requests_mock.Mocker()
    def test_create_event(self, m):

        m.post(self.van.connection.uri + 'events', json=750000984, status_code=204)

        # Test that it doesn't throw and error
        r = self.van.create_event('Canvass 01', 'Can01', '2016-06-01', '2016-06-02',
                                  296199, [259236], publicly_viewable='True',
                                  editable=False)

        self.assertEqual(r, 750000984)

    @requests_mock.Mocker()
    def test_get_event_types(self, m):

        json = [{'eventTypeId': 296199,
                 'name': 'Block Party',
                         'canHaveMultipleShifts': False,
                         'canHaveMultipleLocations': False,
                         'canHaveGoals': False,
                         'canHaveRoleMaximums': False,
                         'canHaveRoleMinimums': False,
                         'canBeRepeatable': False,
                 'roles': [{'roleId': 259236,
                            'name': 'Attendee',
                            'isEventLead': False},
                           {'roleId': 259235,
                            'name': 'Supporter',
                                    'isEventLead': False},
                           {'roleId': 259234,
                            'name': 'Volunteer',
                            'isEventLead': False}],
                 'statuses': [{'statusId': 4, 'name': 'Invited'},
                              {'statusId': 18, 'name': 'Left Msg'},
                              {'statusId': 14, 'name': 'Tentative'},
                              {'statusId': 3, 'name': 'Declined'},
                              {'statusId': 11, 'name': 'Confirmed'},
                              {'statusId': 23, 'name': 'Conf Twice'},
                              {'statusId': 2, 'name': 'Completed'},
                              {'statusId': 15, 'name': 'Walk In'},
                              {'statusId': 6, 'name': 'No Show'},
                              {'statusId': 29, 'name': 'Texted'}],
                 'color': '#7F7F7F',
                 'isAtLeastOneLocationRequired': False,
                 'defaultLocation': None,
                 'isSharedWithMasterCommitteeByDefault': False,
                 'isSharedWithChildCommitteesByDefault': False,
                 'isOnlineActionsAvailable': False}]

        m.get(self.van.connection.uri + 'events/types', json=json)

        expected = ['eventTypeId', 'name', 'canHaveMultipleShifts', 'canHaveMultipleLocations',
                    'canHaveGoals', 'canHaveRoleMaximums', 'canHaveRoleMinimums', 'canBeRepeatable',
                    'roles', 'statuses', 'color', 'isAtLeastOneLocationRequired', 'defaultLocation',
                    'isSharedWithMasterCommitteeByDefault', 'isSharedWithChildCommitteesByDefault',
                    'isOnlineActionsAvailable']

        self.assertTrue(validate_list(expected, self.van.get_event_types()))



"""NGPVAN Saved List Endpoints"""

from parsons.etl.table import Table
from parsons.utilities import cloud_storage
import logging
import uuid
from suds.client import Client

logger = logging.getLogger(__name__)


class SavedLists(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_saved_lists(self, folder_id=None):
        """
        Get saved lists.

        `Args:`
            folder_id: int
                Filter by the id for a VAN folder. If included returns only
                the saved lists in the folder
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('savedLists', params={'folderId': folder_id}))
        logger.info(f'Found {tbl.num_rows} saved lists.')
        return tbl

    def get_saved_list(self, saved_list_id):
        """
        Returns a saved list object.

        `Args:`
            saved_list_id: int
                The saved list id.
        `Returns:`
            dict
        """

        r = self.connection.get_request(f'savedLists/{saved_list_id}')
        logger.info(f'Found saved list {saved_list_id}.')
        return r

    def download_saved_list(self, saved_list_id):
        """
        Download the vanids associated with a saved list.

        `Args:`
            saved_list_id: int
                The saved list id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        ej = ExportJobs(self.connection)
        job = ej.export_job_create(saved_list_id)

        if isinstance(job, tuple):
            return job
        else:
            return Table.from_csv(job['downloadUrl'])

    def upload_saved_list(self, tbl, list_name, folder_id, url_type, id_type='vanid', replace=False,
                          **url_kwargs):
        """
        Upload a saved list. Invalid or unmatched person id records will be ignored. Your api user
        must be shared on the target folder.

        `Args:`
            tbl: parsons.Table
                A parsons table object containing one column of person ids.
            list_name: str
                The saved list name.
            folder_id: int
                The folder id where the list will be stored.
            url_post_type: str
                The cloud file storage to use to post the file. Currently only ``S3``.
            id_type: str
                The primary key type. The options, beyond ``vanid`` are specific to your
                instance of VAN.
            replace: boolean
                Replace saved list if already exists.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type.
                    * S3 requires ``bucket`` argument and, if not stored as env variables
                      ``aws_access_key`` and ``aws_secret_access_key``.
        `Returns:`
            dict
                Upload results information included the number of matched and saved
                records in your list.
        """

        # Move to cloud storage
        file_name = str(uuid.uuid1())
        url = cloud_storage.post_file(tbl, url_type, file_path=file_name + '.zip', **url_kwargs)
        logger.info(f'Table uploaded to {url_type}.')

        # Create XML
        xml = self.connection.soap_client.factory.create('CreateAndStoreSavedListMetaData')
        xml.SavedList._Name = list_name
        xml.DestinationFolder._ID = folder_id
        xml.SourceFile.FileName = file_name + '.csv'
        xml.SourceFile.FileUrl = url
        xml.SourceFile.FileCompression = 'zip'
        xml.Options.OverwriteExistingList = replace

        # Describe file
        file_desc = self.connection.soap_client.factory.create('SeparatedFileFormatDescription')
        file_desc._name = 'csv'
        file_desc.HasHeaderRow = True

        # Only support single column for now
        col = self.connection.soap_client.factory.create('Column')
        col.Name = id_type
        col.RefersTo._Path = f"Person[@PersonIDType=\'{id_type}\']"
        col._Index = '0'

        # VAN errors for this method are not particularly useful or helpful. For that reason, we
        # will check that the folder exists and if the list already exists.
        logger.info('Validating folder id and list name.')
        if folder_id not in [x['folderId'] for x in self.get_folders()]:
            raise ValueError("Folder does not exist or is not shared with API user.")

        if not replace:
            if list_name in [x['name'] for x in self.get_saved_lists(folder_id)]:
                raise ValueError("Saved list already exists. Set to replace argument to True or "
                                 "change list name.")

        # Assemble request
        file_desc.Columns.Column.append(col)
        xml.SourceFile.Format = file_desc

        r = Client.dict(self.connection.soap_client.service.CreateAndStoreSavedList(xml))
        if r:
            logger.info(f"Uploaded {r['ListSize']} records to {r['_Name']} saved list.")
        return r


class Folders(object):

    def __init__(self, van_connection):

        # Some sort of test if the van_connection is not present.

        self.connection = van_connection

    def get_folders(self):
        """
        Get all folders owned or shared with the API user.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('folders'))
        logger.info(f'Found {tbl.num_rows} folders.')
        return tbl

    def get_folder(self, folder_id):
        """
        Get a folder owned by or shared with the API user.

        `Args:`
            folder_id: int
                The folder id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.connection.get_request(f'folders/{folder_id}')
        logger.info(f'Found folder {folder_id}.')
        return r


class ExportJobs(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_export_job_types(self):
        """
        Get export job types

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('exportJobTypes'))
        logger.info(f'Found {tbl.num_rows} export job types.')
        return tbl

    def export_job_create(self, list_id, export_type=4,
                          webhookUrl="https://www.nothing.com"):
        """
        Creates an export job

        Currently, this is only used for exporting saved lists. It is
        recommended that you use the :meth:`saved_list_download` method
        instead.

        `Args:`
            list_id: int
                This is where you should input the list id
            export_type: int
                The export type id, which defines the columns to export
            webhookUrl:
                A webhook to include to notify as to the status of the export
        `Returns:`
            dict
                The export job object
        """

        json = {"savedListId": str(list_id),
                "type": str(export_type),
                "webhookUrl": webhookUrl
                }

        r = self.connection.post_request('exportJobs', json=json)
        logger.info('Retrieved export job.')
        return r

    def get_export_job(self, export_job_id):
        """
        Get an export job.

        `Args:`
            export_job_id: int
                The xxport job id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.connection.get_request(f'exportJobs/{export_job_id}')
        logger.info(f'Found export job {export_job_id}.')
        return r
